@@ -5,7 +5,6 @@ Routines for printing a report.
 from __future__ import print_function, division, absolute_import
 
 import sys
-from collections import namedtuple
 from contextlib import contextmanager
 import textwrap
 from .adapters import BACK, FRONT, PREFIX, SUFFIX, ANYWHERE, LINKED
@@ -14,12 +13,51 @@ from .filters import (NoFilter, PairedNoFilter, TooShortReadFilter, TooLongReadF
 	DiscardTrimmedFilter, DiscardUntrimmedFilter, Demultiplexer, NContentFilter)
 
 
+def safe_divide(numerator, denominator):
+	if numerator is None or not denominator:
+		return 0.0
+	else:
+		return numerator / denominator
+
+
 class Statistics:
-	def __init__(self, n, total_bp1, total_bp2):
+	def __init__(self):
+		"""
+		"""
+		self.n = 0
+		self.total_bp = 0
+		self.total_bp1 = 0
+		self.total_bp2 = 0
+		self.paired = False
+		self.time = 0.01  # CPU time in seconds
+		self.too_short = None
+		self.too_long = None
+		self.written = 0
+		self.written_bp = [0, 0]
+		self.too_many_n = None
+		self.with_adapters = [0, 0]
+		self.quality_trimmed_bp = [0, 0]
+		self.did_quality_trimming = False
+		self.quality_trimmed = 0
+		self.adapter_stats = [None, None]
+		self.adapter_lists = [[], []]
+
+		# These attributes are derived from the ones above
+		self.total_written_bp = 0
+		self.too_short_fraction = 0
+		self.too_long_fraction = 0
+		self.total_written_bp_fraction = 0
+		self.with_adapters_fraction = []
+		self.written_fraction = 0
+		self.quality_trimmed_fraction = 0
+		self.too_many_n_fraction = None
+
+	def collect(self, n, total_bp1, total_bp2, time, modifiers, modifiers2, writers):
 		"""
 		n -- total number of reads
 		total_bp1 -- number of bases in first reads
-		total_bp2 -- number of bases in second reads (set to None for single-end data)
+		total_bp2 -- number of bases in second reads. None for single-end data.
+		time -- CPU time
 		"""
 		self.n = n
 		self.total_bp = total_bp1
@@ -30,21 +68,15 @@ class Statistics:
 			self.paired = True
 			self.total_bp2 = total_bp2
 			self.total_bp += total_bp2
-
-	def collect(self, adapters_pair, time, modifiers, modifiers2, writers):
 		self.time = max(time, 0.01)
-		self.too_short = None
-		self.too_long = None
-		self.written = 0
-		self.written_bp = [0, 0]
-		self.too_many_n = None
+
 		# Collect statistics from writers/filters
 		for w in writers:
-			if isinstance(w, (NoFilter, PairedNoFilter, Demultiplexer)) or isinstance(w.filter, (DiscardTrimmedFilter, DiscardUntrimmedFilter)):
+			if isinstance(w, (NoFilter, PairedNoFilter, Demultiplexer)) or \
+					isinstance(w.filter, (DiscardTrimmedFilter, DiscardUntrimmedFilter)):
 				self.written += w.written
-				if self.n > 0:
-					self.written_fraction = self.written / self.n
-				self.written_bp = self.written_bp[0] + w.written_bp[0], self.written_bp[1] + w.written_bp[1]
+				self.written_bp[0] += w.written_bp[0]
+				self.written_bp[1] += w.written_bp[1]
 			elif isinstance(w.filter, TooShortReadFilter):
 				self.too_short = w.filtered
 			elif isinstance(w.filter, TooLongReadFilter):
@@ -54,9 +86,6 @@ class Statistics:
 		assert self.written is not None
 
 		# Collect statistics from modifiers
-		self.with_adapters = [0, 0]
-		self.quality_trimmed_bp = [0, 0]
-		self.did_quality_trimming = False
 		for i, modifiers_list in [(0, modifiers), (1, modifiers2)]:
 			for modifier in modifiers_list:
 				if isinstance(modifier, QualityTrimmer):
@@ -64,20 +93,19 @@ class Statistics:
 					self.did_quality_trimming = True
 				elif isinstance(modifier, AdapterCutter):
 					self.with_adapters[i] += modifier.with_adapters
-		self.with_adapters_fraction = [ (v / self.n if self.n > 0 else 0) for v in self.with_adapters ]
+					self.adapter_stats[i] = modifier.adapter_statistics
+					self.adapter_lists[i] = modifier.adapters
+
+		# Set the attributes that are derived from the other ones
 		self.quality_trimmed = sum(self.quality_trimmed_bp)
-		self.quality_trimmed_fraction = self.quality_trimmed / self.total_bp if self.total_bp > 0 else 0.0
-
 		self.total_written_bp = sum(self.written_bp)
-		self.total_written_bp_fraction = self.total_written_bp / self.total_bp if self.total_bp > 0 else 0.0
-
-		if self.n > 0:
-			if self.too_short is not None:
-				self.too_short_fraction = self.too_short / self.n
-			if self.too_long is not None:
-				self.too_long_fraction = self.too_long / self.n
-			if self.too_many_n is not None:
-				self.too_many_n_fraction = self.too_many_n / self.n
+		self.written_fraction = safe_divide(self.written, self.n)
+		self.with_adapters_fraction = [safe_divide(v, self.n) for v in self.with_adapters]
+		self.quality_trimmed_fraction = safe_divide(self.quality_trimmed, self.total_bp)
+		self.total_written_bp_fraction = safe_divide(self.total_written_bp, self.total_bp)
+		self.too_short_fraction = safe_divide(self.too_short, self.n)
+		self.too_long_fraction = safe_divide(self.too_long, self.n)
+		self.too_many_n_fraction = safe_divide(self.too_many_n, self.n)
 
 
 ADAPTER_TYPES = {
@@ -104,30 +132,45 @@ def print_error_ranges(adapter_length, error_rate):
 	print()
 
 
-def print_histogram(d, adapter_length, n, error_rate, errors):
+def print_histogram(adapter_statistics, where, adapter, n, gc_content):
 	"""
 	Print a histogram. Also, print the no. of reads expected to be
 	trimmed by chance (assuming a uniform distribution of nucleotides in the reads).
-	d -- a dictionary mapping lengths of trimmed sequences to their respective frequency
+
+	adapter_statistics -- AdapterStatistics object
+	where -- 'front' or 'back'
 	adapter_length -- adapter length
 	n -- total no. of reads.
 	"""
-	h = []
+	if where not in ('front', 'back'):
+		assert ValueError('where must be "front" or "back"')
+	if where == 'front':
+		d = adapter_statistics.lengths_front
+		errors = adapter_statistics.errors_front
+	else:
+		d = adapter_statistics.lengths_back
+		errors = adapter_statistics.errors_back
+
+	match_probabilities = adapter.random_match_probabilities(gc_content=gc_content)
+	print("length", "count", "expect", "max.err", "error counts", sep="\t")
 	for length in sorted(d):
 		# when length surpasses adapter_length, the
 		# probability does not increase anymore
-		estimated = n * 0.25 ** min(length, adapter_length)
-		h.append( (length, d[length], estimated) )
-
-	print("length", "count", "expect", "max.err", "error counts", sep="\t")
-	for length, count, estimate in h:
+		expect = n * match_probabilities[min(len(adapter), length)]
+		count = d[length]
 		max_errors = max(errors[length].keys())
 		errs = ' '.join(str(errors[length][e]) for e in range(max_errors+1))
-		print(length, count, "{0:.1F}".format(estimate), int(error_rate*min(length, adapter_length)), errs, sep="\t")
+		print(
+			length,
+			count,
+			"{0:.1F}".format(expect),
+			int(adapter.max_error_rate*min(length, len(adapter))),
+			errs,
+			sep="\t")
 	print()
 
 
-def print_adjacent_bases(bases, sequence):
+def print_adjacent_bases(bases):
 	"""
 	Print a summary of the bases preceding removed adapter sequences.
 	Print a warning if one of the bases is overrepresented and there are
@@ -168,7 +211,7 @@ def redirect_standard_output(file):
 	sys.stdout = old_stdout
 
 
-def print_report(stats, adapters_pair):
+def print_report(stats, gc_content):
 	"""Print report to standard output."""
 	if stats.n == 0:
 		print("No reads processed! Either your input file is empty or you used the wrong -f/--format parameter.")
@@ -225,9 +268,10 @@ def print_report(stats, adapters_pair):
 
 	warning = False
 	for which_in_pair in (0, 1):
-		for adapter in adapters_pair[which_in_pair]:
-			total_front = sum(adapter.lengths_front.values())
-			total_back = sum(adapter.lengths_back.values())
+		for adapter in stats.adapter_lists[which_in_pair]:
+			adapter_statistics = stats.adapter_stats[which_in_pair][adapter]
+			total_front = sum(adapter_statistics.lengths_front.values())
+			total_back = sum(adapter_statistics.lengths_back.values())
 			total = total_front + total_back
 			where = adapter.where
 			assert where in (ANYWHERE, LINKED) or (where in (BACK, SUFFIX) and total_front == 0) or (where in (FRONT, PREFIX) and total_back == 0)
@@ -239,9 +283,11 @@ def print_report(stats, adapters_pair):
 
 			print("=" * 3, extra + "Adapter", adapter.name, "=" * 3)
 			print()
+
 			if where == LINKED:
-				print("Sequence: {0}...{1}; Type: linked; Length: {2}+{3}; Trimmed: {4} times; Half matches: {5}".
-					format(adapter.front_adapter.sequence,
+				print("Sequence: {0}...{1}; Type: linked; Length: {2}+{3}; "
+					"5' trimmed: {4} times; 3' trimmed: {5} times".format(
+						adapter.front_adapter.sequence,
 						adapter.back_adapter.sequence,
 						len(adapter.front_adapter.sequence),
 						len(adapter.back_adapter.sequence),
@@ -259,36 +305,33 @@ def print_report(stats, adapters_pair):
 				print()
 				print_error_ranges(len(adapter), adapter.max_error_rate)
 				print("Overview of removed sequences (5')")
-				print_histogram(adapter.lengths_front, len(adapter), stats.n, adapter.max_error_rate, adapter.errors_front)
+				print_histogram(adapter_statistics, 'front', adapter, stats.n, gc_content)
 				print()
 				print("Overview of removed sequences (3' or within)")
-				print_histogram(adapter.lengths_back, len(adapter), stats.n, adapter.max_error_rate, adapter.errors_back)
+				print_histogram(adapter_statistics, 'back', adapter, stats.n, gc_content)
 			elif where == LINKED:
 				print()
 				print_error_ranges(len(adapter.front_adapter), adapter.front_adapter.max_error_rate)
 				print_error_ranges(len(adapter.back_adapter), adapter.back_adapter.max_error_rate)
 				print("Overview of removed sequences at 5' end")
-				print_histogram(adapter.front_adapter.lengths_front,
-					len(adapter.front_adapter), stats.n,
-					adapter.front_adapter.max_error_rate,
-					adapter.front_adapter.errors_front)
+				print_histogram(adapter_statistics, 'front',
+					adapter.front_adapter, stats.n, gc_content)
 				print()
 				print("Overview of removed sequences at 3' end")
-				print_histogram(adapter.back_adapter.lengths_back,
-					len(adapter.back_adapter), stats.n,
-					adapter.back_adapter.max_error_rate, adapter.back_adapter.errors_back)
+				print_histogram(adapter_statistics, 'back',
+					adapter.back_adapter, stats.n, gc_content)
 			elif where in (FRONT, PREFIX):
 				print()
 				print_error_ranges(len(adapter), adapter.max_error_rate)
 				print("Overview of removed sequences")
-				print_histogram(adapter.lengths_front, len(adapter), stats.n, adapter.max_error_rate, adapter.errors_front)
+				print_histogram(adapter_statistics, 'front', adapter, stats.n, gc_content)
 			else:
 				assert where in (BACK, SUFFIX)
 				print()
 				print_error_ranges(len(adapter), adapter.max_error_rate)
-				warning = warning or print_adjacent_bases(adapter.adjacent_bases, adapter.sequence)
+				warning = warning or print_adjacent_bases(adapter_statistics.adjacent_bases)
 				print("Overview of removed sequences")
-				print_histogram(adapter.lengths_back, len(adapter), stats.n, adapter.max_error_rate, adapter.errors_back)
+				print_histogram(adapter_statistics, 'back', adapter, stats.n, gc_content)
 
 	if warning:
 		print('WARNING:')
