@@ -6,6 +6,7 @@ from cutadapt.seqio import Sequence
 from cutadapt.adapters import (Adapter, Match, ColorspaceAdapter, FRONT, BACK,
 	parse_braces, LinkedAdapter)
 
+
 def test_issue_52():
 	adapter = Adapter(
 		sequence='GAACTCCAGTCACNNNNN',
@@ -15,7 +16,8 @@ def test_issue_52():
 		read_wildcards=False,
 		adapter_wildcards=True)
 	read = Sequence(name="abc", sequence='CCCCAGAACTACAGTCCCGGC')
-	am = Match(astart=0, astop=17, rstart=5, rstop=21, matches=15, errors=2, front=None, adapter=adapter, read=read)
+	am = Match(astart=0, astop=17, rstart=5, rstop=21, matches=15, errors=2,
+		remove_before=False, adapter=adapter, read=read)
 	assert am.wildcards() == 'GGC'
 	"""
 	The result above should actually be 'CGGC' since the correct
@@ -91,8 +93,7 @@ def test_parse_braces_fail():
 def test_linked_adapter():
 	linked_adapter = LinkedAdapter('AAAA', 'TTTT')
 	sequence = Sequence(name='seq', sequence='AAAACCCCCTTTT')
-	match = linked_adapter.match_to(sequence)
-	trimmed = linked_adapter.trimmed(match)
+	trimmed = linked_adapter.match_to(sequence).trimmed()
 	assert trimmed.name == 'seq'
 	assert trimmed.sequence == 'CCCCC'
 
@@ -107,9 +108,8 @@ def test_info_record():
 		adapter_wildcards=True,
 		name="Foo")
 	read = Sequence(name="abc", sequence='CCCCAGAACTACAGTCCCGGC')
-	am = Match(astart=0, astop=17, rstart=5, rstop=21, matches=15, errors=2, front=None, 
+	am = Match(astart=0, astop=17, rstart=5, rstop=21, matches=15, errors=2, remove_before=False,
 		adapter=adapter, read=read)
-	print(am.get_info_record())
 	assert am.get_info_record() == (
 		"abc",
 		2,
@@ -123,3 +123,19 @@ def test_info_record():
 		'', 
 		''
 	)
+
+
+def test_random_match_probabilities():
+	a = Adapter('A', where=BACK, max_error_rate=0.1)
+	assert a.random_match_probabilities(0.5) == [1, 0.25]
+	assert a.random_match_probabilities(0.2) == [1, 0.4]
+
+	for s in ('ACTG', 'XMWH'):
+		a = Adapter(s, where=BACK, max_error_rate=0.1)
+		assert a.random_match_probabilities(0.5) == [1, 0.25, 0.25**2, 0.25**3, 0.25**4]
+		assert a.random_match_probabilities(0.2) == [1, 0.4, 0.4*0.1, 0.4*0.1*0.4, 0.4*0.1*0.4*0.1]
+
+	a = Adapter('GTCA', where=FRONT, max_error_rate=0.1)
+	assert a.random_match_probabilities(0.5) == [1, 0.25, 0.25**2, 0.25**3, 0.25**4]
+	assert a.random_match_probabilities(0.2) == [1, 0.4, 0.4*0.1, 0.4*0.1*0.4, 0.4*0.1*0.4*0.1]
+
